"""Collection of stastical methods: Transformation and Skill Scores."""

from scipy import stats

__author__ = 'Lukas Leufen, Felix Kleinert'
__date__ = '2019-10-23'

import numpy as np
import xarray as xr
import pandas as pd
from typing import Union, Tuple, Dict, List
from matplotlib import pyplot as plt
import itertools

from mlair.helpers import to_list

Data = Union[xr.DataArray, pd.DataFrame]


def apply_inverse_transformation(data: Data, method: str = "standardise", mean: Data = None, std: Data = None,
                                 max: Data = None, min: Data = None) -> Data:
    """
    Apply inverse transformation for given statistics.

    :param data: transform this data back
    :param method: transformation method (optional)
    :param mean: mean of transformation (optional)
    :param std: standard deviation of transformation (optional)
    :param max: maximum value for min/max transformation (optional)
    :param min: minimum value for min/max transformation (optional)

    :return: inverse transformed data
    """
    if method == 'standardise':  # pragma: no branch
        return standardise_inverse(data, mean, std)
    elif method == 'centre':  # pragma: no branch
        return centre_inverse(data, mean)
    elif method == 'min_max':  # pragma: no branch
        return min_max_inverse(data, min, max)
    elif method == "log":
        return log_inverse(data, mean, std)
    else:
        raise NotImplementedError


def standardise(data: Data, dim: Union[str, int]) -> Tuple[Data, Dict[(str, Data)]]:
    """
    Standardise a xarray.dataarray (along dim) or pandas.DataFrame (along axis) with mean=0 and std=1.

    :param data: data to standardise
    :param dim: name (xarray) or axis (pandas) of dimension which should be standardised
    :return: standardised data, and dictionary with keys method, mean, and standard deviation
    """
    return (data - data.mean(dim)) / data.std(dim), {"mean": data.mean(dim), "std": data.std(dim),
                                                     "method": "standardise"}


def standardise_inverse(data: Data, mean: Data, std: Data) -> Data:
    """
    Apply inverse function of `standardise` on data and therefore vanishes the standardising.

    :param data: standardised data
    :param mean: mean of standardisation
    :param std: standard deviation of transformation

    :return: inverse standardised data
    """
    return data * std + mean


def standardise_apply(data: Data, mean: Data, std: Data) -> Data:
    """
    Apply `standardise` on data using given mean and std.

    :param data: data to transform
    :param mean: mean to use for transformation
    :param std: standard deviation for transformation

    :return: transformed data
    """
    return (data - mean) / std


def centre(data: Data, dim: Union[str, int]) -> Tuple[Data, Dict[(str, Data)]]:
    """
    Centre a xarray.dataarray (along dim) or pandas.DataFrame (along axis) to mean=0.

    :param data: data to centre
    :param dim: name (xarray) or axis (pandas) of dimension which should be centred

    :return: centred data, and dictionary with keys method, and mean
    """
    return data - data.mean(dim), {"mean": data.mean(dim), "method": "centre"}


def centre_inverse(data: Data, mean: Data) -> Data:
    """
    Apply inverse function of `centre` and therefore add given values of mean to data.

    :param data: data to apply inverse centering
    :param mean: mean to use for inverse transformation

    :return: inverted centering transformation data
    """
    return data + mean


def centre_apply(data: Data, mean: Data) -> Data:
    """
    Apply `centre` on data using given mean.

    :param data: data to transform
    :param mean: mean to use for transformation

    :return: transformed data
    """
    return data - mean


def min_max(data: Data, dim: Union[str, int]) -> Tuple[Data, Dict[(str, Data)]]:
    """
    Apply min/max scaling using (x - x_min) / (x_max - x_min). Returned data is in interval [0, 1].

    :param data: data to transform
    :param dim: name (xarray) or axis (pandas) of dimension which should be centred
    :return: transformed data, and dictionary with keys method, min, and max
    """
    d_max = data.max(dim)
    d_min = data.min(dim)
    return (data - d_min) / (d_max - d_min), {"min": d_min, "max": d_max, "method": "min_max"}


def min_max_inverse(data: Data, min: Data, max: Data) -> Data:
    """
    Apply inverse transformation of `min_max` scaling.

    :param data: data to apply inverse scaling
    :param min: minimum value to use for min/max scaling
    :param max: maximum value to use for min/max scaling
    :return: inverted min/max scaled data
    """
    return data * (max - min) + min


def min_max_apply(data: Data, min: Data, max: Data) -> Data:
    """
    Apply `min_max` scaling with given minimum and maximum.

    :param data: data to apply scaling
    :param min: minimum value to use for min/max scaling
    :param max: maximum value to use for min/max scaling
    :return: min/max scaled data
    """
    return (data - min) / (max - min)


def log(data: Data, dim: Union[str, int]) -> Tuple[Data, Dict[(str, Data)]]:
    """
    Apply logarithmic transformation (and standarization) to data. This method first uses the logarithm for
    transformation and second applies the `standardise` method additionally. A logarithmic function numpy's log1p is
    used (`res = log(1+x)`) instead of the pure logarithm to be applicable to values of 0 too.

    :param data: transform this data
    :param dim: name (xarray) or axis (pandas) of dimension which should be transformed
    :return: transformed data, and option dictionary with keys method, mean, and std
    """
    transformed_standardized, opts = standardise(np.log1p(data), dim)
    opts.update({"method": "log"})
    return transformed_standardized, opts


def log_inverse(data: Data, mean: Data, std: Data) -> Data:
    """
    Apply inverse log transformation (therefore exponential transformation). Because `log` is using `np.log1p` this
    method is based on the equivalent method `np.exp1m`. Data are first rescaled using `standardise_inverse` and then
    given to the exponential function.

    :param data: apply inverse log transformation on this data
    :param mean: mean of the standarization
    :param std: std of the standarization
    :return: inverted data
    """
    data_rescaled = standardise_inverse(data, mean, std)
    return np.expm1(data_rescaled)


def log_apply(data: Data, mean: Data, std: Data) -> Data:
    """
    Apply numpy's log1p on given data. Further information can be found in description of `log` method.

    :param data: transform this data
    :param mean: mean of the standarization
    :param std: std of the standarization
    :return: transformed data
    """
    return standardise_apply(np.log1p(data), mean, std)


def mean_squared_error(a, b):
    """Calculate mean squared error."""
    return np.square(a - b).mean()


class SkillScores:
    r"""
    Calculate different kinds of skill scores.

    Skill score on MSE:
        Calculate skill score based on MSE for given forecast, reference and observations.

        .. math::

            \text{SkillScore} = 1 - \frac{\text{MSE(obs, for)}}{\text{MSE(obs, ref)}}

        To run:

        .. code-block:: python

            skill_scores = SkillScores(None).general_skill_score(data, observation_name, forecast_name, reference_name)

    Competitive skill score:
        Calculate skill scores to highlight differences between forecasts. This skill score is also based on the MSE.
        Currently required forecasts are CNN, OLS and persi, as well as the observation obs.

        .. code-block:: python

            skill_scores_class = SkillScores(internal_data)  # must contain columns CNN, OLS, persi and obs.
            skill_scores = skill_scores_class.skill_scores(window_lead_time=3)

    Skill score according to Murphy:
        Follow climatological skill score definition of Murphy (1988). External data is data from another time period
        than the internal data set on initialisation. In other terms, this should be the train and validation data
        whereas the external data is the test data. This sounds perhaps counter-intuitive, but if a skill score is
        evaluated to a model to another, this must be performed on test data set. Therefore, for this case the foreign
        data is test.

        .. code-block:: python

            skill_scores_class = SkillScores(external_data)  # must contain columns obs and CNN.
            skill_scores_clim = skill_scores_class.climatological_skill_scores(internal_data, window_lead_time=3)

    """
    models_default = ["cnn", "persi", "ols"]

    def __init__(self, external_data: Data, models=None, observation_name="obs"):
        """Set internal data."""
        self.external_data = external_data
        self.models = self.set_model_names(models)
        self.observation_name = observation_name

    def set_model_names(self, models: List[str]) -> List[str]:
        """Either use given models or use defaults."""
        if models is None:
            models = self.models_default
        return self._reorder(models)

    @staticmethod
    def _reorder(model_list: List[str]) -> List[str]:
        """Set elements persi and obs at the very end of given list."""
        for element in ["persi", "obs"]:
            if element in model_list:
                model_list.append(model_list.pop(model_list.index(element)))
        return model_list

    def get_model_name_combinations(self):
        """Return all combinations of two models as tuple and string."""
        combinations = list(itertools.combinations(self.models, 2))
        combination_strings = [f"{first}-{second}" for (first, second) in combinations]
        return combinations, combination_strings

    def skill_scores(self, window_lead_time: int) -> pd.DataFrame:
        """
        Calculate skill scores for all combinations of model names.

        :param window_lead_time: length of forecast steps

        :return: skill score for each comparison and forecast step
        """
        ahead_names = list(range(1, window_lead_time + 1))
        combinations, combination_strings = self.get_model_name_combinations()
        skill_score = pd.DataFrame(index=combination_strings)
        for iahead in ahead_names:
            data = self.external_data.sel(ahead=iahead)
            skill_score[iahead] = [self.general_skill_score(data,
                                                            forecast_name=first,
                                                            reference_name=second,
                                                            observation_name=self.observation_name)
                                   for (first, second) in combinations]
        return skill_score

    def climatological_skill_scores(self, internal_data: Data, window_lead_time: int,
                                    forecast_name: str) -> xr.DataArray:
        """
        Calculate climatological skill scores according to Murphy (1988).

        Calculate all CASES I - IV and terms [ABC][I-IV]. Internal data has to be set by initialisation, external data
        is part of parameters.

        :param internal_data: internal data
        :param window_lead_time: interested time step of forecast horizon to select data
        :param forecast_name: name of the forecast to use for this calculation (must be available in `data`)

        :return: all CASES as well as all terms
        """
        ahead_names = list(range(1, window_lead_time + 1))

        all_terms = ['AI', 'AII', 'AIII', 'AIV', 'BI', 'BII', 'BIV', 'CI', 'CIV', 'CASE I', 'CASE II', 'CASE III',
                     'CASE IV']
        skill_score = xr.DataArray(np.full((len(all_terms), len(ahead_names)), np.nan), coords=[all_terms, ahead_names],
                                   dims=['terms', 'ahead'])

        for iahead in ahead_names:
            data = internal_data.sel(ahead=iahead)

            skill_score.loc[["CASE I", "AI", "BI", "CI"], iahead] = np.stack(self._climatological_skill_score(
                data, mu_type=1, forecast_name=forecast_name, observation_name=self.observation_name).values.flatten())

            skill_score.loc[["CASE II", "AII", "BII"], iahead] = np.stack(self._climatological_skill_score(
                data, mu_type=2, forecast_name=forecast_name, observation_name=self.observation_name).values.flatten())

            if self.external_data is not None:
                external_data = self.external_data.sel(ahead=iahead, type=[self.observation_name])
                skill_score.loc[["CASE III", "AIII"], iahead] = np.stack(self._climatological_skill_score(
                    data, mu_type=3, forecast_name=forecast_name, observation_name=self.observation_name,
                    external_data=external_data).values.flatten())

                skill_score.loc[["CASE IV", "AIV", "BIV", "CIV"], iahead] = np.stack(self._climatological_skill_score(
                    data, mu_type=4, forecast_name=forecast_name, observation_name=self.observation_name,
                    external_data=external_data).values.flatten())

        return skill_score

    def _climatological_skill_score(self, internal_data, observation_name, forecast_name, mu_type=1,
                                    external_data=None):
        kwargs = {"external_data": external_data} if external_data is not None else {}
        return self.__getattribute__(f"skill_score_mu_case_{mu_type}")(internal_data, observation_name, forecast_name,
                                                                       **kwargs)

    def general_skill_score(self, data: Data, forecast_name: str, reference_name: str,
                            observation_name: str = None) -> np.ndarray:
        r"""
        Calculate general skill score based on mean squared error.

        :param data: internal data containing data for observation, forecast and reference
        :param observation_name: name of observation
        :param forecast_name: name of forecast
        :param reference_name: name of reference

        :return: skill score of forecast
        """
        if observation_name is None:
            observation_name = self.observation_name
        data = data.dropna("index")
        observation = data.sel(type=observation_name)
        forecast = data.sel(type=forecast_name)
        reference = data.sel(type=reference_name)
        mse = mean_squared_error
        skill_score = 1 - mse(observation, forecast) / mse(observation, reference)
        return skill_score.values

    @staticmethod
    def skill_score_pre_calculations(data: Data, observation_name: str, forecast_name: str) -> Tuple[np.ndarray,
                                                                                                     np.ndarray,
                                                                                                     np.ndarray,
                                                                                                     Data,
                                                                                                     Dict[str, Data]]:
        """
        Calculate terms AI, BI, and CI, mean, variance and pearson's correlation and clean up data.

        The additional information on mean, variance and pearson's correlation (and the p-value) are returned as
        dictionary with the corresponding keys mean, sigma, r and p.

        :param data: internal data to use for calculations
        :param observation_name: name of observation
        :param forecast_name: name of forecast

        :returns: Terms AI, BI, and CI, internal data without nans and mean, variance, correlation and its p-value
        """
        data = data.sel(type=[observation_name, forecast_name]).drop("ahead")
        data = data.dropna("index")

        mean = data.mean("index")
        sigma = np.sqrt(data.var("index"))
        r, p = stats.pearsonr(*[data.sel(type=n).values.flatten() for n in [forecast_name, observation_name]])

        AI = np.array(r ** 2)
        BI = ((r - (sigma.sel(type=forecast_name, drop=True) / sigma.sel(type=observation_name,
                                                                         drop=True))) ** 2).values
        CI = (((mean.sel(type=forecast_name, drop=True) - mean.sel(type=observation_name, drop=True)) / sigma.sel(
            type=observation_name, drop=True)) ** 2).values

        suffix = {"mean": mean, "sigma": sigma, "r": r, "p": p}
        return AI, BI, CI, data, suffix

    def skill_score_mu_case_1(self, internal_data, observation_name, forecast_name):
        """Calculate CASE I."""
        AI, BI, CI, data, _ = self.skill_score_pre_calculations(internal_data, observation_name, forecast_name)
        skill_score = np.array(AI - BI - CI)
        return pd.DataFrame({"skill_score": [skill_score], "AI": [AI], "BI": [BI], "CI": [CI]}).to_xarray().to_array()

    def skill_score_mu_case_2(self, internal_data, observation_name, forecast_name):
        """Calculate CASE II."""
        AI, BI, CI, data, suffix = self.skill_score_pre_calculations(internal_data, observation_name, forecast_name)
        monthly_mean = self.create_monthly_mean_from_daily_data(data)
        data = xr.concat([data, monthly_mean], dim="type")
        sigma = suffix["sigma"]
        sigma_monthly = np.sqrt(monthly_mean.var())
        r, p = stats.pearsonr(*[data.sel(type=n).values.flatten() for n in [observation_name, observation_name + "X"]])
        AII = np.array(r ** 2)
        BII = ((r - sigma_monthly / sigma.sel(type=observation_name, drop=True)) ** 2).values
        skill_score = np.array((AI - BI - CI - AII + BII) / (1 - AII + BII))
        return pd.DataFrame({"skill_score": [skill_score], "AII": [AII], "BII": [BII]}).to_xarray().to_array()

    def skill_score_mu_case_3(self, internal_data, observation_name, forecast_name, external_data=None):
        """Calculate CASE III."""
        AI, BI, CI, data, suffix = self.skill_score_pre_calculations(internal_data, observation_name, forecast_name)
        mean, sigma = suffix["mean"], suffix["sigma"]
        AIII = (((external_data.mean().values - mean.sel(type=observation_name, drop=True)) / sigma.sel(
            type=observation_name, drop=True)) ** 2).values
        skill_score = np.array((AI - BI - CI + AIII) / (1 + AIII))
        return pd.DataFrame({"skill_score": [skill_score], "AIII": [AIII]}).to_xarray().to_array()

    def skill_score_mu_case_4(self, internal_data, observation_name, forecast_name, external_data=None):
        """Calculate CASE IV."""
        AI, BI, CI, data, suffix = self.skill_score_pre_calculations(internal_data, observation_name, forecast_name)
        monthly_mean_external = self.create_monthly_mean_from_daily_data(external_data, index=data.index)
        data = xr.concat([data, monthly_mean_external], dim="type")
        mean, sigma = suffix["mean"], suffix["sigma"]
        mean_external = monthly_mean_external.mean()
        sigma_external = np.sqrt(monthly_mean_external.var())
        r_mu, p_mu = stats.pearsonr(
            *[data.sel(type=n).values.flatten() for n in [observation_name, observation_name + "X"]])
        AIV = np.array(r_mu ** 2)
        BIV = ((r_mu - sigma_external / sigma.sel(type=observation_name, drop=True)) ** 2).values
        CIV = (((mean_external - mean.sel(type=observation_name, drop=True)) / sigma.sel(type=observation_name,
                                                                                         drop=True)) ** 2).values
        skill_score = np.array((AI - BI - CI - AIV + BIV + CIV) / (1 - AIV + BIV + CIV))
        return pd.DataFrame(
            {"skill_score": [skill_score], "AIV": [AIV], "BIV": [BIV], "CIV": CIV}).to_xarray().to_array()

    @staticmethod
    def create_monthly_mean_from_daily_data(data, columns=None, index=None):
        """
        Calculate average for each month and save as daily values with flag 'X'.

        :param data: data to average
        :param columns: columns to work on (all columns from given data are used if empty)
        :param index: index of returned data (index of given data is used if empty)

        :return: data containing monthly means in daily resolution
        """
        if columns is None:
            columns = data.type.values
        if index is None:
            index = data.index
        coordinates = [index, [v + "X" for v in list(columns)]]
        empty_data = np.full((len(index), len(columns)), np.nan)
        monthly_mean = xr.DataArray(empty_data, coords=coordinates, dims=["index", "type"])
        mu = data.groupby("index.month").mean()

        for month in mu.month:
            monthly_mean[monthly_mean.index.dt.month == month, :] = mu[mu.month == month].values.flatten()

        return monthly_mean


class KolmogorovZurbenkoBaseClass:

    def __init__(self, df, wl, itr, is_child=False, filter_dim="window"):
        """
        It create the variables associate with the Kolmogorov-Zurbenko-filter.

        Args:
            df(pd.DataFrame, None): time series of a variable
            wl(list of int): window length
            itr(list of int): number of iteration
        """
        self.df = df
        self.filter_dim = filter_dim
        self.wl = to_list(wl)
        self.itr = to_list(itr)
        if abs(len(self.wl) - len(self.itr)) > 0:
            raise ValueError("Length of lists for wl and itr must agree!")
        self._isChild = is_child
        self.child = self.set_child()
        self.type = type(self).__name__

    def set_child(self):
        if len(self.wl) > 1:
            return KolmogorovZurbenkoBaseClass(None, self.wl[1:], self.itr[1:], True, self.filter_dim)
        else:
            return None

    def kz_filter(self, df, m, k):
        pass

    def spectral_calc(self):
        df_start = self.df
        kz = self.kz_filter(df_start, self.wl[0], self.itr[0])
        filtered = self.subtract(df_start, kz)
        # case I: no child avail -> return kz and remaining
        if self.child is None:
            return [kz, filtered]
        # case II: has child -> return current kz and all child results
        else:
            self.child.df = filtered
            kz_next = self.child.spectral_calc()
            return [kz] + kz_next

    @staticmethod
    def subtract(minuend, subtrahend):
        try:  # pandas implementation
            return minuend.sub(subtrahend, axis=0)
        except AttributeError:  # general implementation
            return minuend - subtrahend

    def run(self):
        return self.spectral_calc()

    def transfer_function(self):
        m = self.wl[0]
        k = self.itr[0]
        omega = np.linspace(0.00001, 0.15, 5000)
        return omega, (np.sin(m * np.pi * omega) / (m * np.sin(np.pi * omega))) ** (2 * k)

    def omega_null(self, alpha=0.5):
        a = np.sqrt(6) / np.pi
        b = 1 / (2 * np.array(self.itr))
        c = 1 - alpha ** b
        d = np.array(self.wl) ** 2 - alpha ** b
        return a * np.sqrt(c / d)

    def period_null(self, alpha=0.5):
        return 1. / self.omega_null(alpha)

    def period_null_days(self, alpha=0.5):
        return self.period_null(alpha) / 24.

    def plot_transfer_function(self, fig=None, name=None):
        if fig is None:
            fig = plt.figure()
        omega, transfer_function = self.transfer_function()
        if self.child is not None:
            transfer_function_child = self.child.plot_transfer_function(fig)
        else:
            transfer_function_child = transfer_function * 0
        plt.semilogx(omega, transfer_function - transfer_function_child,
                     label="m={:3.0f}, k={:3.0f}, T={:6.2f}d".format(self.wl[0],
                                                                     self.itr[0],
                                                                     self.period_null_days()))
        plt.axvline(x=self.omega_null())
        if not self._isChild:
            locs, labels = plt.xticks()
            plt.xticks(locs, np.round(1. / (locs * 24), 1))
            plt.xlim([0.00001, 0.15])
            plt.legend()
            if name is None:
                plt.show()
            else:
                plt.savefig(name)
        else:
            return transfer_function


class KolmogorovZurbenkoFilterMovingWindow(KolmogorovZurbenkoBaseClass):

    def __init__(self, df, wl: Union[list, int], itr: Union[list, int], is_child=False, filter_dim="window",
                 method="mean", percentile=0.5):
        """
        It create the variables associate with the KolmogorovZurbenkoFilterMovingWindow class.

        Args:
            df(pd.DataFrame, xr.DataArray): time series of a variable
            wl: window length
            itr: number of iteration
        """
        self.valid_methods = ["mean", "percentile", "median", "max", "min"]
        if method not in self.valid_methods:
            raise ValueError("Method '{}' is not supported. Please select from [{}].".format(
                method, ", ".join(self.valid_methods)))
        else:
            self.method = method
            if percentile > 1 or percentile < 0:
                raise ValueError("Percentile must be in range [0, 1]. Given was {}!".format(percentile))
            else:
                self.percentile = percentile
        super().__init__(df, wl, itr, is_child, filter_dim)

    def set_child(self):
        if len(self.wl) > 1:
            return KolmogorovZurbenkoFilterMovingWindow(self.df, self.wl[1:], self.itr[1:], is_child=True,
                                                        filter_dim=self.filter_dim, method=self.method,
                                                        percentile=self.percentile)
        else:
            return None

    def kz_filter(self, df, wl, itr):
        """
        It passes the low frequency time series.

        Args:
             wl(int): a window length
             itr(int): a number of iteration
        """
        df_itr = df.__deepcopy__()
        try:
            kwargs = {"min_periods": 1,
                      "center": True,
                      self.filter_dim: wl}
            iter_vars = df_itr.coords["variables"].values
            for var in iter_vars:
                df_itr_var = df_itr.sel(variables=[var]).chunk()
                for _ in np.arange(0, itr):
                    rolling = df_itr_var.rolling(**kwargs)
                    if self.method == "median":
                        df_mv_avg_tmp = rolling.median()
                    elif self.method == "percentile":
                        df_mv_avg_tmp = rolling.quantile(self.percentile)
                    elif self.method == "max":
                        df_mv_avg_tmp = rolling.max()
                    elif self.method == "min":
                        df_mv_avg_tmp = rolling.min()
                    else:
                        df_mv_avg_tmp = rolling.mean()
                    df_itr_var = df_mv_avg_tmp.compute()
                df_itr = df_itr.drop_sel(variables=var).combine_first(df_itr_var)
            return df_itr
        except ValueError:
            raise ValueError
