__author__ = "Lukas Leufen"
__date__ = '2021-02-'

from functools import reduce, partial

from mlair.model_modules import AbstractModelClass
from mlair.helpers import select_from_dict
from mlair.model_modules.loss import var_loss, custom_loss

import keras


class FCN_64_32_16(AbstractModelClass):
    """
    A customised model 4 Dense layers (64, 32, 16, window_lead_time), where the last layer is the output layer depending
    on the window_lead_time parameter.
    """

    def __init__(self, input_shape: list, output_shape: list):
        """
        Sets model and loss depending on the given arguments.

        :param input_shape: list of input shapes (expect len=1 with shape=(window_hist, station, variables))
        :param output_shape: list of output shapes (expect len=1 with shape=(window_forecast))
        """

        assert len(input_shape) == 1
        assert len(output_shape) == 1
        super().__init__(input_shape[0], output_shape[0])

        # settings
        self.activation = keras.layers.PReLU

        # apply to model
        self.set_model()
        self.set_compile_options()
        self.set_custom_objects(loss=self.compile_options['loss'])

    def set_model(self):
        """
        Build the model.
        """
        x_input = keras.layers.Input(shape=self._input_shape)
        x_in = keras.layers.Flatten()(x_input)
        x_in = keras.layers.Dense(64, name="Dense_64")(x_in)
        x_in = self.activation()(x_in)
        x_in = keras.layers.Dense(32, name="Dense_32")(x_in)
        x_in = self.activation()(x_in)
        x_in = keras.layers.Dense(16, name="Dense_16")(x_in)
        x_in = self.activation()(x_in)
        x_in = keras.layers.Dense(self._output_shape, name="Dense_output")(x_in)
        out_main = self.activation()(x_in)
        self.model = keras.Model(inputs=x_input, outputs=[out_main])

    def set_compile_options(self):
        self.optimizer = keras.optimizers.adam(lr=1e-2)
        self.compile_options = {"loss": [keras.losses.mean_squared_error], "metrics": ["mse", "mae"]}


class FCN(AbstractModelClass):
    """
    A customisable fully connected network (64, 32, 16, window_lead_time), where the last layer is the output layer depending
    on the window_lead_time parameter.
    """

    _activation = {"relu": keras.layers.ReLU, "tanh": partial(keras.layers.Activation, "tanh"),
                   "sigmoid": partial(keras.layers.Activation, "sigmoid"),
                   "linear": partial(keras.layers.Activation, "linear"),
                   "selu": partial(keras.layers.Activation, "selu")}
    _initializer = {"selu": keras.initializers.lecun_normal()}
    _optimizer = {"adam": keras.optimizers.adam, "sgd": keras.optimizers.SGD}
    _regularizer = {"l1": keras.regularizers.l1, "l2": keras.regularizers.l2, "l1_l2": keras.regularizers.l1_l2}
    _requirements = ["lr", "beta_1", "beta_2", "epsilon", "decay", "amsgrad", "momentum", "nesterov", "l1", "l2"]

    def __init__(self, input_shape: list, output_shape: list, activation="relu", activation_output="linear",
                 optimizer="adam", n_layer=1, n_hidden=10, regularizer=None, dropout=None, **kwargs):
        """
        Sets model and loss depending on the given arguments.

        :param input_shape: list of input shapes (expect len=1 with shape=(window_hist, station, variables))
        :param output_shape: list of output shapes (expect len=1 with shape=(window_forecast))
        """

        assert len(input_shape) == 1
        assert len(output_shape) == 1
        super().__init__(input_shape[0], output_shape[0])

        # settings
        self.activation = self._set_activation(activation)
        self.activation_output = self._set_activation(activation_output)
        self.optimizer = self._set_optimizer(optimizer, **kwargs)
        self.layer_configuration = (n_layer, n_hidden)
        self._update_model_name()
        self.kernel_initializer = self._initializer.get(activation, "glorot_uniform")
        self.kernel_regularizer = self._set_regularizer(regularizer, **kwargs)
        self.dropout = self._set_dropout(dropout)

        # apply to model
        self.set_model()
        self.set_compile_options()
        self.set_custom_objects(loss=custom_loss([keras.losses.mean_squared_error, var_loss]), var_loss=var_loss)

    def _set_activation(self, activation):
        try:
            return self._activation.get(activation.lower())
        except KeyError:
            raise AttributeError(f"Given activation {activation} is not supported in this model class.")

    def _set_optimizer(self, optimizer, **kwargs):
        try:
            opt_name = optimizer.lower()
            opt = self._optimizer.get(opt_name)
            opt_kwargs = {}
            if opt_name == "adam":
                opt_kwargs = select_from_dict(kwargs, ["lr", "beta_1", "beta_2", "epsilon", "decay", "amsgrad"])
            elif opt_name == "sgd":
                opt_kwargs = select_from_dict(kwargs, ["lr", "momentum", "decay", "nesterov"])
            return opt(**opt_kwargs)
        except KeyError:
            raise AttributeError(f"Given optimizer {optimizer} is not supported in this model class.")

    def _set_regularizer(self, regularizer, **kwargs):
        if regularizer is None:
            return regularizer
        try:
            reg_name = regularizer.lower()
            reg = self._regularizer.get(reg_name)
            reg_kwargs = {}
            if reg_name in ["l1", "l2"]:
                reg_kwargs = select_from_dict(kwargs, reg_name, remove_none=True)
                if reg_name in reg_kwargs:
                    reg_kwargs["l"] = reg_kwargs.pop(reg_name)
            elif reg_name == "l1_l2":
                reg_kwargs = select_from_dict(kwargs, ["l1", "l2"], remove_none=True)
            return reg(**reg_kwargs)
        except KeyError:
            raise AttributeError(f"Given regularizer {regularizer} is not supported in this model class.")

    @staticmethod
    def _set_dropout(dropout):
        if dropout is None:
            return dropout
        assert 0 <= dropout < 1
        return dropout

    def _update_model_name(self):
        n_layer, n_hidden = self.layer_configuration
        n_input = str(reduce(lambda x, y: x * y, self._input_shape))
        n_output = str(self._output_shape)
        self.model_name += "_".join(["", n_input, *[f"{n_hidden}" for _ in range(n_layer)], n_output])

    def set_model(self):
        """
        Build the model.
        """
        x_input = keras.layers.Input(shape=self._input_shape)
        x_in = keras.layers.Flatten()(x_input)
        n_layer, n_hidden = self.layer_configuration
        for layer in range(n_layer):
            x_in = keras.layers.Dense(n_hidden, kernel_initializer=self.kernel_initializer,
                                      kernel_regularizer=self.kernel_regularizer)(x_in)
            x_in = self.activation()(x_in)
            if self.dropout is not None:
                x_in = keras.layers.Dropout(self.dropout)(x_in)
        x_in = keras.layers.Dense(self._output_shape)(x_in)
        out = self.activation_output()(x_in)
        self.model = keras.Model(inputs=x_input, outputs=[out])

    def set_compile_options(self):
        self.compile_options = {"loss": [custom_loss([keras.losses.mean_squared_error, var_loss])],
                                "metrics": ["mse", "mae", var_loss]}
